import random

from robot import *
import environment as env

class Goal:
    def __init__(self, pos, game, time):
        self.pos = pos
        self.game = game
        self.time = time # "randomized" time available to achieve goal
    def makeCommand(self):
        """When a Goal returns None, it is time to find another goal"""
        # TODO: Look at robots around us to determine bid
        game = self.game
        x,y = game.me.pos
        goalpos = self.pos
        if (x,y)==goalpos: return None
        if random.randint(0, self.time+1)==0: return None
        dist = game.dist
        choices = [
            ('N', dist(goalpos, (x, y+1))),
            ('S', dist(goalpos, (x, y-1))),
            ('E', dist(goalpos, (x+1, y))),
            ('W', dist(goalpos, (x-1, y)))]
        mindist = min([c[1] for c in choices if c[1]!=-1])
        choices = [c for c in choices if c[1]==mindist]
        if not choices:
            env.error("Nowhere to go in goal from %s to %s" % ((x,y), goalpos))
            return noopCommand()
        return MoveCommand(direction=random.choice(choices)[0])

def dropPackages(game):
    """Returns command (without bidding) that drops all packages
    to be dropped at this location.
    Returns None if no packages"""
    # TODO: Is this too slow? Should *our* packages
    # be indexed by destination? (There can be a lot)
    droplist = []
    mypos = game.me.pos
    for id in game.me.packages:
        if game.package_index.get(id).destination==mypos:
            droplist.append(id)
    if droplist: return DropCommand(droplist)
    else: return None

def pickupPackages(game, maxweight):
    """Returns a command (without bidding) that
    takes some packages at this location.
    If no packages to take, returns None.
    'maxweight' is the max weight we have left to carry
    """
    # Find the packages that we can carry
    packs = []
    for id in game.package_index.getPos(game.me.pos):
        pack = game.package_index.get(id)
        if pack.weight<=maxweight: packs.append((id, pack))
    if not packs: return None
    # Take the location of a random package. Then pick packages nearest to this pos
    some_pos = random.choice(packs)[1].destination
    for (id, pack) in packs:
        pack.dist = game.dist(some_pos, pack.destination)
    packs.sort(lambda pack1, pack2: cmp(pack1[1].dist, pack2[1].dist))

    picklist = []    
    while maxweight>0 and packs:
        (id, pack) = packs.pop()
        w = pack.weight 
        if w <= maxweight: picklist.append(id); maxweight -= w
        
    if picklist: return PickCommand(picklist)
    else: return None

class Candidate:
    """Candidate for next goal square"""
    def __init__(self, type, pos):
        self.pos = pos
        self.type = type # "DEST", "HOME", "PACK"
    
def destinationSum(id_list, game):
    """id_list is list of package IDs.
    Returns map  pos -> "total weight for that destination"
    """
    result = {}
    for id in id_list:
        pack = game.package_index.get(id)
        dest = pack.destination
        if not dest: continue
        result[dest] = result.get(dest, 0) + pack.weight
    return result

def randomCandidate(candidates, max, game):
    while candidates:
        cand = candidates.pop()
        cand.dist = game.dist(cand.pos, game.me.pos)
        if cand.dist<=max: return cand
    return None

def simpleStrategy(game):
    if len(game.robots)>1 and random.randint(0,3)==0:
        # Make a single random move, to
        # be able to eventually pass a blocking robot. Or something.
        # (But not if we are alone in the world)
        return None 
    
    # Destinations for my packages
    destCandidates = []
    dest_squares = destinationSum(id_list=game.me.packages, game=game)
    for (dest, weight) in dest_squares.iteritems(): # TODO: Use the weight for something
        destCandidates.append(Candidate(type="DEST", pos=dest))

    # Packages squares - only count packages that we can carry
    packCandidates = []
    cw = game.me.carried_weight(package_index = game.package_index)
    spare_capacity = game.maxcap - cw
    for (pos, id_list) in game.package_index.pos_map.iteritems():
        for id in id_list:
            weight = game.package_index.get(id)
            if weight is not None and weight<=spare_capacity:
                # There's at least one that we can carry
                packCandidates.append(Candidate(type="PACK", pos=pos))
                break

    # Home squares
    homeCandidates = []
    for home in game.board.homes:
        homeCandidates.append(Candidate(type="HOME", pos=home))

    # Now find the next goal square
    max = game.money # TODO: game.money has not been updated
    cand = None
    if random.randint(0,1):
        random.shuffle(destCandidates)
        cand = randomCandidate(candidates=destCandidates, max=max, game=game)
        if cand: max = cand.dist # TODO: Implement 'dist'
    if not cand and random.randint(0,1):
        random.shuffle(packCandidates)
        cand = randomCandidate(candidates=packCandidates, max=max, game=game)
        if cand: max = cand.dist
    if not cand:
        random.shuffle(homeCandidates)
        cand = randomCandidate(candidates=homeCandidates, max=max, game=game)
        if cand: max = cand.dist
    all = destCandidates + packCandidates + homeCandidates
    random.shuffle(all)
    while not cand or random.randint(0,1):
        cand2 = randomCandidate(candidates = all, max = max, game = game)
        if not cand2: break
        cand = cand2
        max = cand.dist
    if cand: return Goal(pos=cand.pos, game=game, time=2*cand.dist)
    else: return None

class myrobot1:
    def __init__(self, game):
        self.game = game
        self.goal = None
    def makeGoal(self):
        self.goal = simpleStrategy(game = self.game)
    def makeCommand(self):
        """High level function - Includes any tactics, search, etc."""
        game = self.game
        # TODO: Tell 'game' how much money we spend
        # TODO: Safety tactics first. Avoid water.

        # Very simple tactics implementation follows
        
        # Drop packages, if any:
        command = dropPackages(self.game)
        if command: return command

        # Pickup packages, if any:
        carried = game.me.carried_weight(game.package_index)
        maxweight = game.maxcap - carried
        env.log("Carries %s - weight %d" % (game.me.packages, carried))
        command = pickupPackages(self.game, maxweight=maxweight)
        if command: return command

        # Where are we heading
        if self.goal:
            # We have a goal already
            try: command = self.goal.makeCommand()
            except Exception, e: env.error("Exception: %s"%e) # Guard against bugs in makeCommand
            if command: return command

        self.makeGoal()
        try: command = self.goal and self.goal.makeCommand()
        except Exception, e: env.error("Exception: %s"%e) # Guard against bugs in makeCommand
        if command: return command
        else: return randomWalkCommand(game=game) # Final resort ;-)
