import random, array, sys

from tools import array2, int_array2, set, intlist_to_string
import environment as env

class Board(array2):
    def __init__(self, X, Y, val, homes=None):
        """val : string"""
        array2.__init__(self, X, Y, val=array.array('c', val))
        self.homes = homes
    def getrow(self, y):
        index = y*self.X
        return self.arr[index : index+self.X].tostring()
    def __repr__(self):
        """Note: Prints with north up, unlike message from server"""
        if not self.printable():
            return array2.__repr__(self)
        else:
            rows = [self.getrow(i)[1:-1] for i in range(self.Y-2, 0, -1)]
            return '\n'.join(rows)
    def copy(self):
        return Board(self.X, self.Y, self.arr.tostring())
    def printable(self): return self.X*self.Y <=2500

def board_frame(X,Y):
    top = (X+2)*'#'
    middle = '#' + X*'.' + '#'
    return Board(X+2, Y+2, ''.join([top]+Y*[middle]+[top]))

def read_board(stream):
    X,Y = map(int, stream.readline().split())
    env.log("Reading board: %dx%d" % (X,Y))
    top = (X+2)*'#'
    lines = [top]
    homes = []
    for y in range(1,Y+1):
        line = '#' + stream.readline().strip() + '#'
        lines.append(line)
        x = -1
        try:
            while 1:
                x = line.index('@', x+1)
                homes.append((x,y))
        except ValueError: pass # No more @'s in the line
    lines.append(top)
    init = ''.join(lines)
    board = Board(X+2, Y+2, val=init, homes=homes)
    return board

def make_distance_map(board, pos):
    """Returns an array of distances to (x,y)"""
    # On 'Guido':
    #     0:00:22
    def transl(p):
        # walkable squares become 10000, blocking squares become -1
        if p in '.@': return 10000 
        else: return -1
    # We use 'h' to save memory (an issue for large boards)
    arr = array.array('h', [transl(p) for p in board.arr])
    wid = board.X
    pos = board.pos_to_index(pos)

    arr[pos] = w = 0
    old_list = [pos]
    new_list = []
    while old_list:
        w += 1
        while old_list:
            pos = old_list.pop()
            # Manually "inlined" code
            if w<arr[pos+1]: arr[pos+1] = w; new_list.append(pos+1)
            if w<arr[pos-1]: arr[pos-1] = w; new_list.append(pos-1)
            if w<arr[pos+wid]: arr[pos+wid] = w; new_list.append(pos+wid)
            if w<arr[pos-wid]: arr[pos-wid] = w; new_list.append(pos-wid)
            
        old_list = new_list
        new_list = []
    return int_array2(board.X, board.Y, arr)

def parse_response(line):
    """Example:
    parse_response("#1 X 1 Y 1 E #2 E")
    returns {'1': ['X', '1', 'Y', '1', 'E'], '2': ['E']}"""
    segments = [seg.strip() for seg in line.split('#')]
    if segments[0]: env.error("Error in response line: %s" % line)
    else: del segments[0:1] # remove first, empty part before first '#'
    result = {}
    for seg in segments:
        items = seg.split()
        robot = int(items[0])
        del items[0:1]
        if robot in result:
            env.error("Robot %d occurs multiple in response: %s" % (robot, line))
            result[robot].extend(items)
        else:
            result[robot] = items
    return result

def parse_package_string(package_string):
    nums = package_string.split()
    if len(nums)%4 != 0:
        env.error("Malformed package string: %s" % (package_string,))
        return []
    try: nums = map(int, nums)
    except ValueError:
        env.error("Malformed package string: %s" % (package_string,))
        return []        
    return [tuple(nums[i:i+4]) for i in range(0, len(nums), 4)]

class RobotState:
    def __init__(self, id):
        self.pos = None
        self.id = id
        self.packages = set()
    def update_one(self, item, take, package_index):
        x,y = self.pos or (None,None)
        if   item=='N': self.pos = (x, y+1)
        elif item=='S': self.pos = (x, y-1)
        elif item=='E': self.pos = (x+1, y)
        elif item=='W': self.pos = (x-1, y)
        elif item=='X': self.pos = (take(), y)
        elif item=='Y': self.pos = (x, take())
        elif item=='D':
            id = take()
            package_index.drop(id, pos=self.pos)
            self.packages.remove(id)
        elif item=='P':
            id = take()
            package_index.pick(id, robot=self.id)
            self.packages.insert(id)
        else:
            env.error("Unexpected response command %s" % (item,))
    def update(self, response_items, package_index):
        """for example: response_items=['X', '1', 'Y', '2']"""
        while response_items:
            item = response_items[0].upper()
            del response_items[0:1]
            def take():
                if not response_items:
                    env.error("Malformed response, cannot take()")
                    return None
                try:
                    id = int(response_items[0])
                except ValueError:
                    env.error("Malformed response, not an integer: %s" % (response_items[0],))
                    return None
                del response_items[0:1];
                return id
            self.update_one(item, take, package_index)
    def carried_weight(self, package_index):
        total = 0
        for id in self.packages: total += package_index.get(id).weight
        return total
    def __repr__(self):
        return "<robot %d at %s>" % (self.id, self.pos)

class PackageState:
    # States of package:
    # where: "FREE" or "CARRIED"
    # pos: (x,y); used if "FREE"
    # robot: ID; used if "CARRIED"
    # destination: (x,y) or None
    # weight: n or None
    # Held by robot n, or lying at position pos
    #
    def __init__(self):
        self.pos = None
        self.robot = None
        self.destination = None
        self.weight = None
    def setPos(self, pos):
        self.where = "FREE"
        self.pos = pos
        self.robot = None
    def setRobot(self, robot):
        self.where = "CARRIED"
        self.pos = None
        self.robot = robot
    def __repr__(self):
        try: where = self.where
        except AttributeError: return "<Package where?>"
        if where=="FREE": return "<Package at %s>" % (self.pos,)
        elif where=="CARRIED": return "<Package at robot #%d>" % self.robot
        else: return "<Package where2?>"
                
def ascii_num(num): return `num%10`

def mk_board_w_state(board, state):
    # TODO: This is too slow. Rather, keep a "robot map" and a "board map"
    board = board.copy()
    for (n, robot) in state.robots.items():
        board[robot.pos] = ascii_num(n)
    return board

class PackageIndex:
    def __init__(self):
        self.id_map = { }  # maps id -> PackageState
        self.pos_map = { } # maps pos -> id list
    def pick(self, id, robot):
        try: pack = self.id_map[id]
        except KeyError: self.id_map[id] = pack = PackageState()
        else:
            # Remove the package from its position
            try: self.pos_map[pack.pos].remove(id) # Fast enough with list?
            except ValueError: env.error("Picking up package that is not there")
        # And tell the package that it is now carried
        pack.setRobot(robot)
    def drop(self, id, pos):
        # TODO: If we know the destination of the package, and if
        # that is 'pos', then just clear it from the game!
        try: pack = self.id_map[id]
        except KeyError:
            # Can happen if we entered in the middle of a game
            self.id_map[id] = pack = PackageState()
        if pack.destination==pos:
            # Package no longer exists
            del self.id_map[id]
            return
        pack.setPos(pos)
        self.pos_map.setdefault(pos, [])
        pack_list = self.pos_map[pos]        
        if id in pack_list:
            env.error("Package %d already at %s?!" % (id, pos))
            return
        self.pos_map[pos].append(id)
    def get(self, id): return self.id_map[id]
    def getPos(self, pos): return self.pos_map.get(pos, [])

class Game:
    """State of game, as seen from one robot client
    (as opposed to the global view that a simulator would need)
    board
    myid
    maxcap
    money      - how much money left
    initmoneey - how much money to start with
    robots     : id -> RobotState
    package_index : PackageIndex
    me         - RobotState, (or None if we're not in the list for some strange reason)
    dist_cache : pos -> distarray
    dist_queue - list of "most recent dist lookup pos's"
    """
    def __init__(self, board, myid, maxcap, money):
        self.board = board
        self.myid = myid
        self.maxcap = maxcap
        self.money = self.initmoney = money
        self.robots = { } 
        self.package_index = PackageIndex()
        self.dist_cache = { } # pos -> int_array2
        # The cache uses about 50MB when full - 2 bytes per entry
        self.max_dist_cache = 25000000/(board.X*board.Y+1) 
    def __repr__(self):
        return `mk_board_w_state(self.board, self)`
    def update(self, response):
        """response - The response string from the server"""
        # We make a new robot map, because some robots might have died
        response = parse_response(response)
        new_robots = {}
        for (id, action) in response.items():
            try: old = self.robots[id]
            except KeyError: old = RobotState(id=id)
            old.update(action, package_index=self.package_index)
            new_robots[id] = old
        self.robots = new_robots
        # May become None if we disappeared!
        self.me = self.robots.get(self.myid, None)
    def update_packages(self, package_string):
        """package_string - The "packages at current location" string as sent from the server"""
        mypos = self.me.pos
        parsed_packages = parse_package_string(package_string)
        # To not see packages at this location that are not there:
        self.package_index.pos_map[mypos] = pos_list = []
        for (id, x, y, weight) in parsed_packages:
            st = PackageState()
            st.where="FREE"
            st.pos = mypos
            st.destination = (x,y)
            st.weight = weight
            pos_list.append(id)
            self.package_index.id_map[id] = st
        if self.board[mypos]=='@':
            # No longer a home base, since we know what's there
            try: self.board.homes.remove(mypos)
            except ValueError: pass # happens if mypos is already removed from homes
    def dist(self, pos1, pos2):
        try: return self.dist_cache[pos1][pos2]
        except KeyError: pass
        try: return self.dist_cache[pos2][pos1]
        except KeyError: pass
        dist = make_distance_map(self.board, pos1)
        self.dist_cache[pos1] = dist
        if len(self.dist_cache)>self.max_dist_cache: # TODO: Test with a small number to see it work
            # Erase a random cache element
            pos = random.choice(self.dist_cache.keys())
            del self.dist_cache[pos]
        return dist[pos2]
    def forget_dist(self, pos):
        try: del self.dist_cache[pos]
        except KeyError: pass # Just ignore errors

# Command interface:
# bid - None or an integer
# tostring
# tostring_nobid

class Command: # TODO: Use it
    def __init__(self): self.bid = None
    def tostring(self):
        if self.bid:
            return "%d %s" % (self.bid, self.tostring_nobid())
        else:
            # env.error("Bid defaulting to 1") # TODO: Reinsert
            return "1 " + self.tostring_nobid()
    def __str__(self): return self.tostring()

class MoveCommand(Command):
    def __init__(self, direction):
        Command.__init__(self)
        self.direction = direction
    def tostring_nobid(self): return "Move " + self.direction

class PickCommand(Command):
    def __init__(self, picklist):
        Command.__init__(self)
        self.picklist = picklist
    def tostring_nobid(self):
        return "Pick " + intlist_to_string(self.picklist)

class DropCommand(Command):
    def __init__(self, droplist):
        Command.__init__(self)
        self.droplist = droplist
    def tostring_nobid(self):
        return "Drop " + intlist_to_string(self.droplist)

def randomWalkCommand(game):
    (x,y) = game.me.pos
    choices = ['.']
    if game.board[x,y+1] in '.@': choices.append('N')
    if game.board[x,y-1] in '.@': choices.append('S')
    if game.board[x+1,y] in '.@': choices.append('E')
    if game.board[x-1,y] in '.@': choices.append('W')
    choice = random.choice(choices)
    if choice=='.': return noopCommand()
    else: return MoveCommand(direction = choice)

def noopCommand(): return DropCommand(droplist=[])
        
def run(robot, istream=sys.stdin, ostream=sys.stdout):
    print >>ostream, "Player"
    ostream.flush()
    world = read_board(istream)
    env.log(world)
    myid, maxcap, money = map(int, istream.readline().split())
    env.log("myid=%d, maxcap=%d, money=%d" % (myid, maxcap, money))
    g = Game(board=world, myid=myid, maxcap=maxcap, money=money)
    rob = robot(g)
    while 1:
        line = istream.readline()
        env.log("Response from server: " + line)
        g.update(line)
        env.log(g)
        g.update_packages(istream.readline())
        cmd = rob.makeCommand()
        env.log("My command: %s"% cmd)
        print >>ostream, cmd
        ostream.flush()        

